//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft end-user
// license agreement (EULA) under which you licensed this SOFTWARE PRODUCT.
// If you did not accept the terms of the EULA, you are not authorized to use
// this source code. For a copy of the EULA, please see the LICENSE.RTF on your
// install media.
//
//
//
// Use of this source code is subject to the terms of the Microsoft end-user
// license agreement (EULA) under which you licensed this SOFTWARE PRODUCT.
// If you did not accept the terms of the EULA, you are not authorized to use
// this source code. For a copy of the EULA, please see the LICENSE.RTF on your
// install media.
//

#include <windows.h>
#include <bsp.h>
#include <ethdbg.h>
#include <fmd.h>
#include "loader.h"
#include "fat.h"
#include "sdi.h"
#include "display.h"
//#include "..\lcd\logo_320_240.h"

#define IS_LB 0

void ShowReadProgress(DWORD dwPacketNum);
static void InitDisplay(DWORD LcdType);

// For USB Download function.
extern BOOL UbootReadData (DWORD cbData, LPBYTE pbData);
extern BOOL InitUSB ();
extern void Isr_Init();
extern void USBDisconnect();
extern void USBReconnect();
extern DWORD SDInit(void);
static BOOL MenuLcd(PBOOT_CFG pBootCfg);

char *inet_ntoa(DWORD dwIP);
DWORD inet_addr( char *pszDottedD );
BOOL EbootInitEtherTransport (EDBG_ADDR *pEdbgAddr, LPDWORD pdwSubnetMask,
                              BOOL *pfJumpImg,
                              DWORD *pdwDHCPLeaseTime,
                              UCHAR VersionMajor, UCHAR VersionMinor,
                              char *szPlatformString, char *szDeviceName,
                              UCHAR CPUId, DWORD dwBootFlags);
BOOL EbootEtherReadData (DWORD cbData, LPBYTE pbData);
EDBG_OS_CONFIG_DATA *EbootWaitForHostConnect (EDBG_ADDR *pDevAddr, EDBG_ADDR *pHostAddr);
// End   ***************************************


// Globals
//
DWORD 			EdbgDebugZone;
DWORD			g_ImageType;
MultiBINInfo	g_BINRegionInfo;
PBOOT_CFG		g_pBootCfg;
UCHAR			g_TOC[SECTOR_SIZE];
const PTOC 		g_pTOC = (PTOC)&g_TOC;
DWORD			g_dwImageStartBlock;
DWORD			g_dwTocEntry;
BOOL			g_bBootMediaExist = FALSE;
BOOL			g_bDownloadImage  = TRUE;
BOOL 			g_bWaitForConnect = TRUE;
BOOLEAN			g_bUSBDownload = FALSE;
BOOLEAN			g_bReadFromSDCard = FALSE;
DWORD			ImageSize;
DWORD			g_ExtPartitionSize = 5;  // EBOOTз ˷С

EDBG_ADDR 		g_DeviceAddr; // NOTE: global used so it remains in scope throughout download process
                        // since eboot library code keeps a global pointer to the variable provided.

DWORD			wNUM_BLOCKS;
DWORD           wBLOCK_SIZE;
// External definitions.
//
extern const BYTE ScreenBitmap[];

BYTE OEMGetKey(void);
/*
    @func   void | SpinForever | Halts execution (used in error conditions).
    @rdesc  
    @comm    
    @xref   
*/
static void SpinForever(void)
{
	volatile DWORD i;
    EdbgOutputDebugString("SpinForever...\r\n");

    while(1)
    {    	
#ifdef SUPPORT_LCD_FUNCTION
    	DrawString(STARTX, LINE(13), "SpinForever...", LCD_RED, LCD_BLACK, NORMAL);
		for (i=1000000; i; i--);
    	DrawString(STARTX, LINE(13), "              ", LCD_RED, LCD_BLACK, NORMAL);
		for (i=1000000; i; i--);
#endif        
    }
}
void SystemPowerOff(void)
{
	volatile S3C2440A_IOPORT_REG *pIORegs = (S3C2440A_IOPORT_REG*)OALPAtoVA(S3C2440A_BASE_REG_PA_IOPORT, FALSE);
	pIORegs->GPGCON &= ~(0x3<<24);
	pIORegs->GPGCON |= (0x1<<24);
	pIORegs->GPGUP |= (0x1<<12);
	pIORegs->GPGDAT &= ~(0x1<<12);
	SpinForever();
}


/*
    @func   void | main | Samsung bootloader C routine entry point.
    @rdesc  N/A.
    @comm    
    @xref   
*/
void main(void)
{
    // Clear LEDs.
    //   
    OEMWriteDebugLED(0, 0x3);
	
    // Common boot loader (blcommon) main routine.
    //    
    BootloaderMain();

    // Should never get here.
    // 
    SpinForever();
}


/*
    @func   void | InitDisplay | Initializes the LCD controller and displays a splashscreen image.
    @rdesc  N/A.
    @comm    
    @xref   
*/
static void InitDisplay(DWORD LcdType)
{
    volatile S3C2440A_IOPORT_REG *s2440IOP = (S3C2440A_IOPORT_REG *)OALPAtoVA(S3C2440A_BASE_REG_PA_IOPORT, FALSE);
    volatile S3C2440A_LCD_REG    *s2440LCD = (S3C2440A_LCD_REG *)OALPAtoVA(S3C2440A_BASE_REG_PA_LCD, FALSE);
	unsigned int clkval_calc;  // VCLKƵ.
    DWORD XSize, YSize;
    DWORD VBPD, VFPD, VSPW; // źŵǰʱ,Լ.
    DWORD HBPD, HFPD, HSPW; // 

    // Set up the LCD controller registers to display a power-on bitmap image.
    //
    s2440IOP->GPCUP     = 0xFFFFFFFF;
    s2440IOP->GPCCON    = 0xAAAA5AAA;
                                    
	//s2440IOP->GPEDAT    = 0x1;
    s2440IOP->GPDUP     = 0xFFFFFFFF;
    s2440IOP->GPDCON    = 0xAAAAAAAA; //0xAAAAAAA1; 
    /*
    * GPG4 is lcd power enable pin.set GPG4 high.
    */
    s2440IOP->GPGCON   &= ~(3 << 8);
    s2440IOP->GPGCON   |=  (1 << 8);
    s2440IOP->GPGDAT   |=  (1 << 4);

    /*
     * tft 640*480
     */
    if (LcdType == TFT640_480) {

        clkval_calc = 1;
        XSize = 640;
        YSize = 480;

        VBPD = ((3)&0xff);
        VFPD = ((3)&0xff);
        VSPW = ((2) &0x3f);
        HBPD = ((12)&0x7f);
        HFPD = ((12)&0xff);
        HSPW = ((30)&0xff);


    /* tft 320*240 */
    } else if (LcdType == TFT320_240) {

    	clkval_calc = 7;//(WORD)((float)(S3C2440A_HCLK)/(2.0*5000000)+0.5)-1;
        XSize = 320;
        YSize = 240;
        
        VBPD  =	((3)&0xff);
        VFPD  =	((5)&0xff);
        VSPW  = ((15) &0x3f);
        HBPD  = ((5)&0x7f);
        HFPD  =	((15)&0xff);
        HSPW  =	((8)&0xff);
    	
    } 
    /* tft 800 * 480*/
    else if (LcdType == TFT800_480) {
        clkval_calc = 1;
        XSize = 800;
        YSize = 480;
        
        VBPD  =	((33-1)&0xff);
        VFPD  =	((10-1)&0xff);
        VSPW  = ((2-1) &0x3f);
        HBPD  = ((48-1)&0x7f);
        HFPD  =	((16-1)&0xff);
        HSPW  =	((96-1)&0xff);
        
    }
    /* tft 240 * 320*/
    else if (LcdType == TFT240_320) {
        clkval_calc = 4;
        XSize = 240;
        YSize = 320;
        
        VBPD  =	((3)&0xff);
        VFPD  =	((5)&0xff);
        VSPW  = ((15) &0x3f);
        HBPD  = ((5)&0x7f);
        HFPD  =	((15)&0xff);
        HSPW  =	((8)&0xff);
        
    }
    s2440LCD->LCDCON1   =  (clkval_calc <<  8) |       /* VCLK = HCLK / ((CLKVAL + 1) * 2) -> About 7 Mhz  */
                           (0           <<  7) |       /*(LCD_MVAL    <<  7)  |    */   /* 0 : Each Frame                                   */
                           (3           <<  5) |       /* TFT LCD Pannel                                   */
                           (12          <<  1) |       /* 16bpp Mode                                       */
                           (0           <<  0) ;       /* Disable LCD Output                               */

   	s2440LCD->LCDCON2   =  (VBPD        << 24) |	// VBPD          :   1
	    				   ((YSize-1)   << 14) |	// Virtical Size : 320 - 1
  						   (VFPD        <<  6) |	// VFPD          :   2
   						   (VSPW        <<  0) ;	// VSPW          :   1

   	s2440LCD->LCDCON3   =  (HBPD        << 19) |	// HBPD          :   6
   						   ((XSize-1)   <<  8) |	// HOZVAL_TFT    : 240 - 1
   						   (HFPD        <<  0) ;	// HFPD          :   2

   	s2440LCD->LCDCON4   =  (13          <<  8) |	// MVAL          :  13                              */
   						   (HSPW        <<  0) ;	// HSPW          :   4                              */


    s2440LCD->LCDCON5   =  (0           << 12) |       /* BPP24BL       : LSB valid                        */
                           (1           << 11) |       /* FRM565 MODE   : 5:6:5 Format                     */
#if (BOARD_TYPE==BOARD_P1310)
                           (1           << 10) |       /* INVVCLK       : VCLK Falling Edge                */
                           (0           <<  9) |       /* INVVLINE      : Inverted Polarity                */
#else
                           (0           << 10) |       /* INVVCLK       : VCLK Falling Edge                */
                           (1           <<  9) |       /* INVVLINE      : Inverted Polarity                */
#endif
                           (1           <<  8) |       /* INVVFRAME     : Inverted Polarity                */
                           (0           <<  7) |       /* INVVD         : Normal                           */
                           (0           <<  6) |       /* INVVDEN       : Normal                           */
                           (0           <<  5) |       /* INVPWREN      : Normal                           */
                           (0           <<  4) |       /* INVENDLINE    : Normal                           */
                           (0           <<  3) |       /* PWREN         : Disable PWREN                    */
                           (0           <<  2) |       /* ENLEND        : Disable LEND signal              */
                           (0           <<  1) |       /* BSWP          : Swap Disable                     */
                           (1           <<  0) ;       /* HWSWP         : Swap Enable                      */



    s2440LCD->LCDSADDR1 = ((IMAGE_FRAMEBUFFER_DMA_BASE >> 22)     << 21) | 
                          ((M5D(IMAGE_FRAMEBUFFER_DMA_BASE >> 1)) <<  0);

    s2440LCD->LCDSADDR2 = M5D((IMAGE_FRAMEBUFFER_DMA_BASE + (XSize * YSize * 2)) >> 1);

    s2440LCD->LCDSADDR3 = (((XSize - XSize) / 1) << 11) | (XSize / 1);        

    //s2440LCD->TCONSEL   |= 0x3;
    s2440LCD->LCDINTMSK |= 3;
  	s2440LCD->TCONSEL   &= (~7);
    //    s2440LCD->TCONSEL   |= (0x1<<4);

   	s2440LCD->TPAL     = 0x0; 
   	s2440LCD->TCONSEL &= ~((1<<4) | 1);								// Disable LCC3600, LCP3600
   	s2440IOP->GPGUP	   = s2440IOP->GPGUP  & (~(1<<4)) | (1<<4);		// Pull-up disbale
   	s2440IOP->GPGCON   = s2440IOP->GPGCON & (~(3<<8)) | (3<<8);
        
   	s2440LCD->LCDCON5  = s2440LCD->LCDCON5 & (~(1<<3)) | (1<<3);	// PWREN
   	s2440LCD->LCDCON5  = s2440LCD->LCDCON5 & (~(1<<5)) | (0<<5);	// INVPWREN    
    s2440LCD->LCDCON1  |= 1;


    // Display a bitmap image on the LCD...
    //
    //memcpy((void *)IMAGE_FRAMEBUFFER_UA_BASE, ScreenBitmap, XSize*YSize*2);//LCD_ARRAY_SIZE_TFT_16BIT);
    memset((void *)IMAGE_FRAMEBUFFER_UA_BASE, 0x00, XSize*YSize*2);//LCD_ARRAY_SIZE_TFT_16BIT);

}


/*
    @func   void | SetIP | Accepts IP address from user input.
    @rdesc  N/A.
    @comm    
    @xref   
*/

static void SetIP(PBOOT_CFG pBootCfg)
{
    CHAR   szDottedD[16];   // The string used to collect the dotted decimal IP address.
    USHORT cwNumChars = 0;
    USHORT InChar = 0;

    EdbgOutputDebugString("\r\nEnter new IP address: ");

    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a number or a period, add it to the string.
            //
            if (InChar == '.' || (InChar >= '0' && InChar <= '9')) 
            {
                if (cwNumChars < 16) 
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up.
            //
            else if (InChar == 8) 
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szDottedD[cwNumChars] = '\0';
        pBootCfg->EdbgAddr.dwIP = inet_addr(szDottedD);
    }
}


/*
    @func   void | SetMask | Accepts subnet mask from user input.
    @rdesc  N/A.
    @comm    
    @xref   
*/
static void SetMask(PBOOT_CFG pBootCfg)
{
    CHAR szDottedD[16]; // The string used to collect the dotted masks.
    USHORT cwNumChars = 0;
    USHORT InChar = 0;

    EdbgOutputDebugString("\r\nEnter new subnet mask: ");

    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a number or a period, add it to the string.
            //
            if (InChar == '.' || (InChar >= '0' && InChar <= '9')) 
            {
                if (cwNumChars < 16) 
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up.
            //
            else if (InChar == 8) 
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szDottedD[cwNumChars] = '\0';
        pBootCfg->SubnetMask = inet_addr(szDottedD);
    }
}


/*
    @func   void | SetDelay | Accepts an autoboot delay value from user input.
    @rdesc  N/A.
    @comm    
    @xref   
*/
static void SetDelay(PBOOT_CFG pBootCfg)
{
    CHAR szCount[16];
    USHORT cwNumChars = 0;
    USHORT InChar = 0;

    EdbgOutputDebugString("\r\nEnter maximum number of seconds to delay [1-255]: ");

    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a number or a period, add it to the string.
            //
            if ((InChar >= '0' && InChar <= '9')) 
            {
                if (cwNumChars < 16) 
                {
                    szCount[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up.
            //
            else if (InChar == 8) 
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szCount[cwNumChars] = '\0';
        pBootCfg->BootDelay = atoi(szCount);
        if (pBootCfg->BootDelay > 255)
        {
            pBootCfg->BootDelay = 255;
        } 
        else if (pBootCfg->BootDelay < 1)
        {
            pBootCfg->BootDelay = 1;
        }
    }
}


static ULONG mystrtoul(PUCHAR pStr, UCHAR nBase)
{
    UCHAR nPos=0;
    BYTE c;
    ULONG nVal = 0;
    UCHAR nCnt=0;
    ULONG n=0;

    // fulllibc doesn't implement isctype or iswctype, which are needed by
    // strtoul, rather than including coredll code, here's our own simple strtoul.

    if (pStr == NULL)
        return(0);

    for (nPos=0 ; nPos < strlen(pStr) ; nPos++)
    {
        c = tolower(*(pStr + strlen(pStr) - 1 - nPos));
        if (c >= '0' && c <= '9')
            c -= '0';
        else if (c >= 'a' && c <= 'f')
        {
            c -= 'a';
            c  = (0xa + c);
        }

        for (nCnt = 0, n = 1 ; nCnt < nPos ; nCnt++)
        {
            n *= nBase;
        }
        nVal += (n * c);
    }

    return(nVal);
}


static void CvtMAC(USHORT MacAddr[3], char *pszDottedD ) 
{
    DWORD cBytes;
    char *pszLastNum;
    int atoi (const char *s);
    int i=0;    
    BYTE *p = (BYTE *)MacAddr;

    // Replace the dots with NULL terminators
    pszLastNum = pszDottedD;
    for(cBytes = 0 ; cBytes < 6 ; cBytes++)
    {
        while(*pszDottedD != '.' && *pszDottedD != '\0')
        {
            pszDottedD++;
        }
        if (pszDottedD == '\0' && cBytes != 5)
        {
            // zero out the rest of MAC address
            while(i++ < 6)
            {
                *p++ = 0;
            }
            break;
        }
        *pszDottedD = '\0';
        *p++ = (BYTE)(mystrtoul(pszLastNum, 16) & 0xFF);
        i++;
        pszLastNum = ++pszDottedD;
    }
}


static void SetCS8900MACAddress(PBOOT_CFG pBootCfg)
{
    CHAR szDottedD[24];
    USHORT cwNumChars = 0;
    USHORT InChar = 0;

    memset(szDottedD, '0', 24);

    EdbgOutputDebugString ( "\r\nEnter new MAC address in hexadecimal (hh.hh.hh.hh.hh.hh): ");

    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        InChar = tolower(InChar);
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a hex number or a period, add it to the string.
            //
            if (InChar == '.' || (InChar >= '0' && InChar <= '9') || (InChar >= 'a' && InChar <= 'f')) 
            {
                if (cwNumChars < 17) 
                {
                    szDottedD[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            else if (InChar == 8)       // If it's a backspace, back up.
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }

    EdbgOutputDebugString ( "\r\n");

    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szDottedD[cwNumChars] = '\0';
        CvtMAC(pBootCfg->EdbgAddr.wMAC, szDottedD);

        EdbgOutputDebugString("INFO: MAC address set to: %x:%x:%x:%x:%x:%x\r\n",
                  pBootCfg->EdbgAddr.wMAC[0] & 0x00FF, pBootCfg->EdbgAddr.wMAC[0] >> 8,
                  pBootCfg->EdbgAddr.wMAC[1] & 0x00FF, pBootCfg->EdbgAddr.wMAC[1] >> 8,
                  pBootCfg->EdbgAddr.wMAC[2] & 0x00FF, pBootCfg->EdbgAddr.wMAC[2] >> 8);
    }
    else
    {
        EdbgOutputDebugString("WARNING: SetCS8900MACAddress: Invalid MAC address.\r\n");
    }
}




UINT32 SetBlockPage(void)
{
    CHAR szCount[16];
    USHORT cwNumChars = 0;
    USHORT InChar = 0;
    UINT32 block=0, page=0;
    UINT32 i;
 
    EdbgOutputDebugString("\r\nEnter Block # : ");
 
    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a number or a period, add it to the string.
            //
            if ((InChar >= '0' && InChar <= '9')) 
            {
                if (cwNumChars < 16) 
                {
                    szCount[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up.
            //
            else if (InChar == 8) 
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }
 
    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szCount[cwNumChars] = '\0';
        block = atoi(szCount);
    }
 
    EdbgOutputDebugString("\r\nEnter Page # : ");
    InChar = 0;
    for (i=0; i<16; i++) szCount[i] = '0';
    
    while(!((InChar == 0x0d) || (InChar == 0x0a)))
    {
        InChar = OEMReadDebugByte();
        if (InChar != OEM_DEBUG_COM_ERROR && InChar != OEM_DEBUG_READ_NODATA) 
        {
            // If it's a number or a period, add it to the string.
            //
            if ((InChar >= '0' && InChar <= '9')) 
            {
                if (cwNumChars < 16) 
                {
                    szCount[cwNumChars++] = (char)InChar;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
            // If it's a backspace, back up.
            //
            else if (InChar == 8) 
            {
                if (cwNumChars > 0) 
                {
                    cwNumChars--;
                    OEMWriteDebugByte((BYTE)InChar);
                }
            }
        }
    }
 
    // If it's a carriage return with an empty string, don't change anything.
    //
    if (cwNumChars) 
    {
        szCount[cwNumChars] = '\0';
        page = atoi(szCount);
    }
 
    if (IS_LB) return ((block<<6)|page);
    else       return ((block<<5)|page);
}
 

void PrintPageData(DWORD nMData, DWORD nSData, UINT8* pMBuf, UINT8* pSBuf)
{
    DWORD i;
 
    OALMSG(TRUE, (TEXT("=========================================================")));
    for (i=0;i<nMData;i++)
    {
        if ((i%16)==0)
            OALMSG(TRUE, (TEXT("\r\n 0x%03x |"), i));
        OALMSG(TRUE, (TEXT(" %02x"), pMBuf[i]));
        if ((i%512)==511)
            OALMSG(TRUE, (TEXT("\r\n ------------------------------------------------------- ")));
    }
    for (i=0;i<nSData;i++)
    {
        if ((i%16)==0)
            OALMSG(TRUE, (TEXT("\r\n 0x%03x |"), i));
        OALMSG(TRUE, (TEXT(" %02x"), pSBuf[i]));
    }
    OALMSG(TRUE, (TEXT("\r\n=========================================================")));
}



/*
    @func   BOOL | MainMenu | Manages the Samsung bootloader main menu.
    @rdesc  TRUE == Success and FALSE == Failure.
    @comm    
    @xref   
*/

static BOOL MainMenu(PBOOT_CFG pBootCfg)
{
    BYTE KeySelect = 0;
    BOOL bConfigChanged = FALSE;
    BOOLEAN bDownload = TRUE;
	DWORD i;
	
    while(TRUE)
    {
        KeySelect = 0;

        EdbgOutputDebugString ( "\r\nEthernet Boot Loader Configuration:\r\n\r\n");
        EdbgOutputDebugString ( "0) IP address: %s\r\n",inet_ntoa(pBootCfg->EdbgAddr.dwIP));
        EdbgOutputDebugString ( "1) Subnet mask: %s\r\n", inet_ntoa(pBootCfg->SubnetMask));
        EdbgOutputDebugString ( "2) DHCP: %s\r\n", (pBootCfg->ConfigFlags & CONFIG_FLAGS_DHCP)?"Enabled":"Disabled");
        EdbgOutputDebugString ( "3) Boot delay: %d seconds\r\n", pBootCfg->BootDelay);
        EdbgOutputDebugString ( "4) Reset to factory default configuration\r\n");
        EdbgOutputDebugString ( "5) Startup image: %s\r\n", (g_pBootCfg->ConfigFlags & BOOT_TYPE_DIRECT) ? "LAUNCH EXISTING" : "DOWNLOAD NEW");
        EdbgOutputDebugString ( "6) Program disk image into SmartMedia card: %s\r\n", (pBootCfg->ConfigFlags & TARGET_TYPE_NAND)?"Enabled":"Disabled");
        EdbgOutputDebugString ( "7) Program CS8900 MAC address (%B:%B:%B:%B:%B:%B)\r\n",
                               g_pBootCfg->EdbgAddr.wMAC[0] & 0x00FF, g_pBootCfg->EdbgAddr.wMAC[0] >> 8,
                               g_pBootCfg->EdbgAddr.wMAC[1] & 0x00FF, g_pBootCfg->EdbgAddr.wMAC[1] >> 8,
                               g_pBootCfg->EdbgAddr.wMAC[2] & 0x00FF, g_pBootCfg->EdbgAddr.wMAC[2] >> 8);
        EdbgOutputDebugString ( "8) Kernel Debugger: %s\r\n", (g_pBootCfg->ConfigFlags & CONFIG_FLAGS_DEBUGGER) ? "ENABLED" : "DISABLED");
        EdbgOutputDebugString ( "9) Format Boot Media for BinFS\r\n");

        // N.B: we need this option here since BinFS is really a RAM image, where you "format" the media
        // with an MBR. There is no way to parse the image to say it's ment to be BinFS enabled.
        EdbgOutputDebugString ( "F) Low-level format the Smart Media card\r\n");
        EdbgOutputDebugString ( "A) Erase All block. \r\n");
        EdbgOutputDebugString ( "D) Download image now\r\n");
        EdbgOutputDebugString ( "L) LAUNCH existing Boot Media image\r\n");
        EdbgOutputDebugString ( "U) DOWNLOAD image now(USB)\r\n");
        EdbgOutputDebugString ( "R) Read Configuration \r\n");
        EdbgOutputDebugString ( "W) Write Configuration Right Now\r\n");
        EdbgOutputDebugString ( "-------------------------------------------\r\n");       
        EdbgOutputDebugString ( "C) Clean Hive register: %s\r\n", \
            (pBSPArgs->bSysHiveClean)?("TRUE"):("FALSE") );
        //EdbgOutputDebugString ( "c) Clean USER Hive register: %s\r\n", \
        //    (pBSPArgs->bUserHiveClean)?("TRUE"):("FALSE") );
        EdbgOutputDebugString ( "s) DOWNLOAD nk.bin image from SD card\r\n");
        EdbgOutputDebugString ( "S) DOWNLOAD eboot.bin image from SD card\r\n");		
        EdbgOutputDebugString ( "M) DOWNLOAD logo.bin image from SD card\r\n");
        EdbgOutputDebugString ( "N) Print Nand flash.\r\n");	
		EdbgOutputDebugString ( "O) update bios.\r\n");	
		EdbgOutputDebugString ( "P) power off.\r\n");	
        EdbgOutputDebugString ( "T) Test.\r\n");		        
        EdbgOutputDebugString ( "\r\nEnter your selection: ");

        while (! ( ( (KeySelect >= '0') && (KeySelect <= '9') ) ||
                   ( (KeySelect == 'C') || (KeySelect == 'c') ) ||            
                   ( (KeySelect == 'D') || (KeySelect == 'd') ) ||
                   ( (KeySelect == 'F') || (KeySelect == 'f') ) ||
                   ( (KeySelect == 'L') || (KeySelect == 'l') ) ||
                   ( (KeySelect == 'U') || (KeySelect == 'u') ) ||
                   ( (KeySelect == 'R') || (KeySelect == 'r') ) ||
                   ( (KeySelect == 'W') || (KeySelect == 'w') ) ||
                   ( (KeySelect == 'S') || (KeySelect == 's') ) ||                  
                   ( (KeySelect == 'N') || (KeySelect == 'n') ) ||                                    
                   ( (KeySelect == 'A') || (KeySelect == 'a') ) ||                                                       
                   ( (KeySelect == 'M') || (KeySelect == 'm') ) ||                   
                   ( (KeySelect == 'T') || (KeySelect == 't') ) ||                   
                   ( (KeySelect == 'P') || (KeySelect == 'p') ) ||                                      
                   ( (KeySelect == 'O') || (KeySelect == 'o') ) ||                                      
                   ( (KeySelect == 'X') || (KeySelect == 'x') ) ))
        {
            KeySelect = OEMReadDebugByte();
        }

        EdbgOutputDebugString ( "%c\r\n", KeySelect);

        switch(KeySelect)
        {
        case '0':           // Change IP address.
            SetIP(pBootCfg);
            pBootCfg->ConfigFlags &= ~CONFIG_FLAGS_DHCP;   // clear DHCP flag
            bConfigChanged = TRUE;
            break;
        case '1':           // Change subnet mask.
            SetMask(pBootCfg);
            bConfigChanged = TRUE;
            break;
        case '2':           // Toggle static/DHCP mode.
            pBootCfg->ConfigFlags = (pBootCfg->ConfigFlags ^ CONFIG_FLAGS_DHCP);
            bConfigChanged = TRUE;
            break;
        case '3':           // Change autoboot delay.
            SetDelay(pBootCfg);
            bConfigChanged = TRUE;
            break;
        case '4':           // Reset the bootloader configuration to defaults.
            OALMSG(TRUE, (TEXT("Resetting default TOC...\r\n")));
            TOC_Init(DEFAULT_IMAGE_DESCRIPTOR, (IMAGE_TYPE_RAMIMAGE|IMAGE_TYPE_BINFS), 0, 0, 0);
            if ( !TOC_Write() ) {
                OALMSG(OAL_WARN, (TEXT("TOC_Write Failed!\r\n")));
            }
            OALMSG(TRUE, (TEXT("...TOC complete\r\n")));
            break;
        case '5':           // Toggle download/launch status.
            pBootCfg->ConfigFlags = (pBootCfg->ConfigFlags ^ BOOT_TYPE_DIRECT);
            bConfigChanged = TRUE;
            break;
        case '6':           // Toggle image storage to Smart Media.
            pBootCfg->ConfigFlags = (pBootCfg->ConfigFlags ^ TARGET_TYPE_NAND);
            bConfigChanged = TRUE;
            break;
        case '7':           // Configure Crystal CS8900 MAC address.
            SetCS8900MACAddress(pBootCfg);
            bConfigChanged = TRUE;
            break;
        case '8':           // Toggle KD
            g_pBootCfg->ConfigFlags = (g_pBootCfg->ConfigFlags ^ CONFIG_FLAGS_DEBUGGER);
            g_bWaitForConnect = (g_pBootCfg->ConfigFlags & CONFIG_FLAGS_DEBUGGER) ? TRUE : FALSE;
            bConfigChanged = TRUE;
            continue;
            break;            
        case '9':
            // format the boot media for BinFS
            // N.B: this does not destroy our OEM reserved sections (TOC, bootloaders, etc)
            if ( !g_bBootMediaExist ) {
				OALMSG(OAL_ERROR, (TEXT("ERROR: BootMonitor: boot media does not exist.\r\n")));
                continue;
            }
            // N.B: format offset by # of reserved blocks,
            // decrease the ttl # blocks available by that amount.
            if ( !BP_LowLevelFormat( g_dwImageStartBlock,
                                     //NAND_BLOCK_CNT - g_dwImageStartBlock, // nandflash޸Ĺˡ
									 wNUM_BLOCKS - g_dwImageStartBlock,
                                     0) )
            {
                OALMSG(OAL_ERROR, (TEXT("ERROR: BootMonitor: Low-level boot media format failed.\r\n")));
                continue;
            }
            break;
        case 'F':
        case 'f':
            // low-level format
            // N.B: this erases images, BinFs, FATFS, user data, etc.
            // However, we don't format Bootloaders & TOC bolcks; use JTAG for this.
            if ( !g_bBootMediaExist ) {
				OALMSG(OAL_ERROR, (TEXT("ERROR: BootMonitor: boot media does not exist.\r\n")));
                continue;
            } else {
                SectorInfo si;

                // to keep bootpart off of our reserved blocks we must mark it as bad, reserved & read-only
                si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
                si.bBadBlock    = BADBLOCKMARK;
                si.dwReserved1  = 0xffffffff;
                si.wReserved2   = 0xffff;

				// Ҫbootл.
                //OALMSG(TRUE, (TEXT("Reserving Blocks [0x%x - 0x%x] ...\r\n"), 0, IMAGE_START_BLOCK-1));
                //for (i = 0; i < IMAGE_START_SECTOR; i++) {
                OALMSG(TRUE, (TEXT("Reserving Blocks [0x%x - 0x%x] ...\r\n"), 0, g_dwImageStartBlock-1));
                for (i = 0; i < BLOCK_TO_SECTOR(g_dwImageStartBlock); i++) {
                    FMD_WriteSector(i, NULL, &si, 1);
                }
                OALMSG(TRUE, (TEXT("...reserve complete.\r\n")));

                //OALMSG(TRUE, (TEXT("Low-level format Blocks [0x%x - 0x%x] ...\r\n"), IMAGE_START_BLOCK, wNUM_BLOCKS-1));
                //for (i = IMAGE_START_BLOCK; i < wNUM_BLOCKS; i++) {
                OALMSG(TRUE, (TEXT("Low-level format Blocks [0x%x - 0x%x] ...\r\n"), g_dwImageStartBlock, wNUM_BLOCKS-1));
                for (i = g_dwImageStartBlock; i < wNUM_BLOCKS; i++) {
                //for (i = IMAGE_START_BLOCK; i < NAND_BLOCK_CNT; i++) {
                    FMD_EraseBlock(i);
                }
                OALMSG(TRUE, (TEXT("...erase complete.\r\n")));
            } break;
        case 'A':           // Download? Yes.
        case 'a':
            if ( !g_bBootMediaExist ) {
				OALMSG(OAL_ERROR, (TEXT("ERROR: BootMonitor: boot media does not exist.\r\n")));
                continue;
            } else {
                //OALMSG(TRUE, (TEXT("Low-level format Blocks [0x%x - 0x%x] ...\r\n"), IMAGE_START_BLOCK, wNUM_BLOCKS-1));
                //for (i = IMAGE_START_BLOCK; i < wNUM_BLOCKS; i++) {
                OALMSG(TRUE, (TEXT("Low-level format Blocks [0x%x - 0x%x] ...\r\n"), 1, wNUM_BLOCKS-1));
                for (i = 1; i < wNUM_BLOCKS; i++) {
                //for (i = IMAGE_START_BLOCK; i < NAND_BLOCK_CNT; i++) {
                    FMD_EraseBlock(i);
                }
                OALMSG(TRUE, (TEXT("...erase complete.\r\n")));
            } break;
            goto MENU_DONE;

        case 'D':           // Download? Yes.
        case 'd':
            bDownload = TRUE;
            goto MENU_DONE;
        case 'L':           // Download? No.
        case 'l':
            bDownload = FALSE;
            goto MENU_DONE;
        case 'R':
        case 'r':
			TOC_Read();
			TOC_Print();
            // TODO
            break;
        case 'U':           // Download? No.
        case 'u':            
            bConfigChanged = TRUE;
			g_bUSBDownload = TRUE;
            bDownload = TRUE;
            goto MENU_DONE;
			
        case 'W':           // Configuration Write
        case 'w':
            if (!TOC_Write())
            {
                OALMSG(OAL_WARN, (TEXT("WARNING: MainMenu: Failed to store updated eboot configuration in flash.\r\n")));
            }
            else
            {
                OALMSG(OAL_INFO, (TEXT("Successfully Written\r\n")));
                bConfigChanged = FALSE;
            }
            break;
        case 'C':
        case 'c':
            pBSPArgs->bSysHiveClean = (pBSPArgs->bSysHiveClean)?FALSE:TRUE;
            pBSPArgs->bUserHiveClean = (pBSPArgs->bUserHiveClean)?FALSE:TRUE;
            break;
            
        case 's':
        case 'S':
		case 'm':
		case 'M':
		{
			LONG ret;
			ret = SDInit();
			if (ret == SD_API_STATUS_DEVICE_NOT_CONNECTED) {
				DrawString(8, 16, "[SD] card not found", LCD_RED, LCD_BLACK, TRANSP);
			} else if (ret) {
				DrawString(8, 16, "[SD] fail to init ", LCD_RED, LCD_BLACK, TRANSP);
			} else {
				DrawString(8, 16, "[SD] init done     ", LCD_BLACK, LCD_WHITE, TRANSP);	// ok		
			}
		
			if (!fatLoad(0)){
				DrawString(8, 32, "[FAT] fail to read", LCD_RED, LCD_BLACK, TRANSP);
				break;
			}else DrawString(8, 32, "[FAT] mounted     ", LCD_BLACK, LCD_WHITE, TRANSP); // ok
			
			if (KeySelect=='s') 
                ImageSize = fatOpen("nk.bin");//file name is not unicode...
			else if (KeySelect=='s')			
                ImageSize = fatOpen("eboot.bin");//file name is not unicode...
			else if (KeySelect=='m' || KeySelect=='M') {
#if 0
                DWORD   i;
                SectorInfo si;
                
                ImageSize = fatOpen("logo.bin");//file name is not unicode...
                if (ImageSize == -1)    RETAILMSG(1, (L"failed to open logo.bin"));
                
                SDCardReadData(ImageSize, IMAGE_FRAMEBUFFER_UA_BASE);
                for (i=LOGO_BLOCK; i<LOGO_BLOCK+LOGO_BLOCK_SIZE; i++) {
                    RETAILMSG(1, (L"write logo to block:%u.\r\n", i));
                    FMD_EraseBlock(i);
                    if (!WriteBlock(i, IMAGE_FRAMEBUFFER_UA_BASE + wBLOCK_SIZE*(i-LOGO_BLOCK), NULL))
                        RETAILMSG(1, (L"failed to write logo to block:%u.", i));
                }

                si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
                si.bBadBlock    = BADBLOCKMARK;
                si.dwReserved1  = 0xffffffff;
                si.wReserved2   = 0xffff;

                for (i = 0; i < BLOCK_TO_SECTOR(g_dwImageStartBlock); i++) {
                    FMD_WriteSector(i, NULL, &si, 1);
                }
                RETAILMSG(1, (L"write logo done."));
                SpinForever();
#endif                
            }

			
			if (ImageSize == -1){
				DrawString(8, 48, "[FILE] .bin not found      ", LCD_RED, LCD_BLACK, TRANSP);				
				break;  
			}else DrawString(8, 48, "[FILE] opened. wait reading", LCD_BLACK, LCD_WHITE, TRANSP); // ok

			//fatRead((void*)0x80200000, ImageSize);

			g_bReadFromSDCard	= TRUE;
			bConfigChanged = TRUE;
			bDownload = TRUE;

			goto MENU_DONE;
            break;
        }
        case 'N':
        case 'n':
            {
                #define READ_LB_NAND_M_SIZE 2048
                #define READ_LB_NAND_S_SIZE 64
                #define READ_SB_NAND_M_SIZE 512
                #define READ_SB_NAND_S_SIZE 16
 
                UINT8  pMBuf[READ_LB_NAND_M_SIZE], pSBuf[READ_LB_NAND_S_SIZE];
                UINT32 rslt;
                UINT32 nMainloop, nSpareloop;
 
                rslt = SetBlockPage();
                OALMSG(TRUE, (TEXT("\r\n rlst : 0x%x\r\n"), rslt));
                memset(pMBuf, 0xff, READ_LB_NAND_M_SIZE);
                memset(pSBuf, 0xff, READ_LB_NAND_S_SIZE);
 
                if (IS_LB)
                {
                    nMainloop  = READ_LB_NAND_M_SIZE;
                    nSpareloop = READ_LB_NAND_S_SIZE;
 
                    //RAW_LB_ReadSector(rslt, pMBuf, pSBuf);
                    //FMD_LB_ReadSector(rslt, pMBuf, pSBuf, 1);
                }
                else
                {
                    nMainloop  = READ_SB_NAND_M_SIZE;
                    nSpareloop = READ_SB_NAND_S_SIZE;
 
                    //RAW_SB_ReadSector(rslt, pMBuf, pSBuf);
                    FMD_ReadSector(rslt, pMBuf, (SectorInfo*)pSBuf, 1);
                }
 
                PrintPageData(nMainloop, nSpareloop, pMBuf, pSBuf);
            }
            break;            
	  case 'o':
	  case 'O':
	  {
	  	    unsigned int length, i;
        	char * add = (char*)0xA2000000;
            SectorInfo si;

            RETAILMSG(1, (L"ready to receive bios image from COM...send now\r\n"));
            
            /* ó */
        	((char*)&length)[0] = OEMBlockReadDebugByte();
        	((char*)&length)[1] = OEMBlockReadDebugByte();
        	((char*)&length)[2] = OEMBlockReadDebugByte();
        	((char*)&length)[3] = OEMBlockReadDebugByte();

        	i = length - 4;

        	while(i--) {
        		*add++ = OEMBlockReadDebugByte();
        	}
            
            RETAILMSG(1, (L"start write bios.\r\n"));

            // TODO: дblock0
            FMD_EraseBlock(0);
            add = (char*)0xA2000000;
            for (i=0; i<256; i++) {
                FMD_WriteSector(i, add, NULL, 1);
                add += 512;
                if ( (DWORD)add - 0xA2000000 > length)
                    break;
            }

            // to keep bootpart off of our reserved blocks we must mark it as bad, reserved & read-only
            si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
            si.bBadBlock    = BADBLOCKMARK;
            si.dwReserved1  = 0xffffffff;
            si.wReserved2   = 0xffff;

			// Ҫbootл.
            OALMSG(TRUE, (TEXT("Reserving Blocks [0x%x - 0x%x] ...\r\n"), 0, g_dwImageStartBlock-1));
            for (i = 0; i < BLOCK_TO_SECTOR(1); i++) {
                FMD_WriteSector(i, NULL, &si, 1);
            }
            OALMSG(TRUE, (TEXT("...reserve complete.\r\n")));

	  }
			break;
	  case 'T':
	  case 't':
	  	{

			char buf[1024];
			unsigned i, j=20*32;

            
            
#if 0						
            SectorInfo Si;
            RETAILMSG(1, (L"\r\nsec:%u\r\n", j));            
            
			FMD_ReadSector(j, buf, &Si, 1);            
            for (i=0; i<512; i++) {
				if (i%32 == 0)	RETAILMSG(1, (L"\r\n"));
				if (i%4 == 0)	RETAILMSG(1, (L" "));
				RETAILMSG(1, (L"%02X", (UCHAR)buf[i]));
			}
			RETAILMSG(1, (L"bad:%02X oem:%02X\r\n", (UCHAR)Si.bBadBlock, 
			(UCHAR)Si.bOEMReserved));
            
            break;

			RETAILMSG(1, (L"\r\nread block start\r\n"));			
			for (j=0;j<10;j++) {
				for(i=0; i<32*1024*1024/512;i++) {
					FMD_ReadSector(i, buf, NULL, 1);
				}
			}
			RETAILMSG(1, (L"end\r\n"));			
#endif			


#if 1
			if (SDInit()) break;

			SDReadSector(0, 1, buf);
			for (i=0; i<512; i++) {
				if (i%32 == 0)	RETAILMSG(1, (L"\r\n"));
				if (i%4 == 0)	RETAILMSG(1, (L" "));
				RETAILMSG(1, (L"%02X", (UCHAR)buf[i]));
			}
			RETAILMSG(1, (L"\r\n"));

			SDReadSector(1, 1, buf);
			for (i=0; i<512; i++) {
				if (i%32 == 0)	RETAILMSG(1, (L"\r\n"));
				if (i%4 == 0)	RETAILMSG(1, (L" "));
				RETAILMSG(1, (L"%02X", (UCHAR)buf[i]));
			}
			RETAILMSG(1, (L"\r\n"));

			SDReadSector(0, 2, buf);
			for (i=0; i<1024; i++) {
				if (i%32 == 0)	RETAILMSG(1, (L"\r\n"));
				if (i%4 == 0)	RETAILMSG(1, (L" "));
				RETAILMSG(1, (L"%02X", (UCHAR)buf[i]));
			}
#else
{
	FATFS fatfs;
	DIR dir;				/* Directory object */
	FIL file1, file2;		/* File object */
	DWORD br;
	//f_mount(0, fs);
	char buff[512];
	DWORD res;	
	int i,j;
	
	f_mount(0, &fatfs);
	res=f_open(&file1, "NK.bin", FA_READ);
	RETAILMSG(1, (L"f_open() return:%x\r\n", res));
	res=f_read(&file1, buff, 512, &br);
	RETAILMSG(1, (L"f_read() return:%x\r\n", res));
	for (i=0; i<512; i++) {
		if (i%32 == 0)	RETAILMSG(1, (L"\r\n"));
		if (i%4 == 0)	RETAILMSG(1, (L" "));
		RETAILMSG(1, (L"%02X", (UCHAR)buff[i]));
	}
	RETAILMSG(1, (L"\r\n"));
	
}

#endif

#if 0
			RETAILMSG(1, (L"\r\nread block start\r\n"));			
			for (i=0; i<2048*30; i++)
				if (SDReadSector(i, 1, buf)) break;
			RETAILMSG(1, (L"read block end:\r\n"));			

#endif			
	  	}
		break;


        case 'p':
        case 'P':
			SystemPowerOff();
            break;

        default:
            break;
        }
    }

MENU_DONE:

    // If eboot settings were changed by user, save them to flash.
    //
    if (bConfigChanged && !TOC_Write())
    {
        OALMSG(OAL_WARN, (TEXT("WARNING: MainMenu: Failed to store updated bootloader configuration to flash.\r\n")));
    }

    return(bDownload);    
}


/*
    @func   BOOL | OEMPlatformInit | Initialize the Samsung SMD2440 platform hardware.
    @rdesc  TRUE = Success, FALSE = Failure.
    @comm    
    @xref   
*/
int OEMPlatformInit(void)
{
    ULONG BootDelay;
    UINT8 KeySelect;
    UINT32 dwStartTime, dwPrevTime, dwCurrTime;
    BOOLEAN bResult = FALSE;
    FlashInfo flashInfo;

    OALMSG(OAL_FUNC, (TEXT("+OEMPlatformInit.\r\n")));

    EdbgOutputDebugString("Microsoft Windows CE Bootloader for the Samsung SMDK2440 Version %d.%d Built %s\r\n\r\n", 
                          EBOOT_VERSION_MAJOR, EBOOT_VERSION_MINOR, __DATE__);


    // Initialize the display.
    //
    InitDisplay(TFT_LCD_TYPE);
	//DispOneColor(LCD_WHITE);

    // Initialize the BSP args structure.
    //
    memset(pBSPArgs, 0, sizeof(BSP_ARGS));
    pBSPArgs->header.signature       = OAL_ARGS_SIGNATURE;
    pBSPArgs->header.oalVersion      = OAL_ARGS_VERSION;
    pBSPArgs->header.bspVersion      = BSP_ARGS_VERSION;
    pBSPArgs->kitl.flags             = OAL_KITL_FLAGS_ENABLED | OAL_KITL_FLAGS_VMINI;
    pBSPArgs->kitl.devLoc.IfcType    = Internal;
    pBSPArgs->kitl.devLoc.BusNumber  = 0;
    pBSPArgs->kitl.devLoc.LogicalLoc = BSP_BASE_REG_PA_CS8900A_IOBASE;

#if 1 // SDȡʱҲҪIsr_Init()ʼ.
	// USB D+ Low Signal 
	USBDisconnect();

	// For USB Download functon
	//
	if (!InitUSB())
	{
		DEBUGMSG(1, (TEXT("OEMPlatformInit: Failed to initialize USB.\r\n")));
		return(FALSE);
	}
	Isr_Init();

   	{	
		volatile int i;
		for(i=10000; i != 0; i--);
		//	OALMSG(TRUE, (TEXT("\r\nWait for 10ms %d "),i));
	}
	// USB D+ High Signal
	USBReconnect();
#endif

    // This should not change unless reserved blocks are added/removed;
    // made global to do the calc only once.
    g_dwImageStartBlock = IMAGE_START_BLOCK;

    // Try to initialize the boot media block driver and BinFS partition.
    //
    if ( !BP_Init((LPBYTE)BINFS_RAM_START, BINFS_RAM_LENGTH, NULL, NULL, NULL) )
    {
        OALMSG(OAL_WARN, (TEXT("WARNING: OEMPlatformInit failed to initialize Boot Media.\r\n")));
        g_bBootMediaExist = FALSE;
    }
    else
        g_bBootMediaExist = TRUE;

    // Get flash info
    if (!FMD_GetInfo(&flashInfo)) {
        OALMSG(1, (L"ERROR: BLFlashDownload: "
            L"FMD_GetInfo call failed\r\n"
        ));
    }
	wNUM_BLOCKS = flashInfo.dwNumBlocks;
    wBLOCK_SIZE = flashInfo.dwBytesPerBlock;
	RETAILMSG(1, (TEXT("wNUM_BLOCKS : %d wBLOCK_SIZE : %d \r\n"), wNUM_BLOCKS, wBLOCK_SIZE));
    // Ҫbootֻ.
	RETAILMSG(1, (L"Searching Image Start Block..."));	
    for(; g_dwImageStartBlock<flashInfo.dwNumBlocks;g_dwImageStartBlock++) {
		if(FMD_GetBlockStatus(g_dwImageStartBlock)==0x3) { // bad , oemreserved.
			RETAILMSG(1, (L"(Skip %u)", g_dwImageStartBlock));
		} else {
			RETAILMSG(1, (L" Done! Block=%u\r\n", g_dwImageStartBlock));
			break;
		}
    }

    // Try to retrieve TOC (and Boot config) from boot media
    //
    if ( !TOC_Read( ) ) 
    {
        // use default settings
        TOC_Init(DEFAULT_IMAGE_DESCRIPTOR, (IMAGE_TYPE_RAMIMAGE), 0, 0, 0);
    }

    // Display boot message - user can halt the autoboot by pressing any key on the serial terminal emulator.
    //
    BootDelay = g_pBootCfg->BootDelay;

    if (g_pBootCfg->ConfigFlags & BOOT_TYPE_DIRECT)
    {
        OALMSG(TRUE, (TEXT("Press [ENTER] to launch image stored on boot media, or [SPACE] to enter boot monitor.\r\n")));
        OALMSG(TRUE, (TEXT("\r\nInitiating image launch in %d seconds. "),BootDelay--));

    }
    else
    {
        OALMSG(TRUE, (TEXT("Press [ENTER] to download image stored on boot media, or [SPACE] to enter boot monitor.\r\n")));
        OALMSG(TRUE, (TEXT("\r\nInitiating image download in %d seconds. "),BootDelay--));
    }
#ifdef SUPPORT_LCD_FUNCTION
    DrawString(STARTX, 0, "Press [ENTER] to enter bios.", LCD_WHITE, LCD_BLACK, TRANSP);
    DrawNum(STARTX, 16, (BootDelay+1), LCD_RED, LCD_BLACK, TRANSP);
#endif

    dwStartTime = OEMEthGetSecs();
    dwPrevTime  = dwStartTime;
    dwCurrTime  = dwStartTime;
    KeySelect   = 0;

    // Allow the user to break into the bootloader menu.
    //
    while((dwCurrTime - dwStartTime) < g_pBootCfg->BootDelay)
    {
        KeySelect = OEMReadDebugByte();
        if ((KeySelect == 0x20) || (KeySelect == 0x0d))
            break;

#ifdef SUPPORT_LCD_FUNCTION
{
	    volatile S3C2440A_IOPORT_REG *s2440IOP = (S3C2440A_IOPORT_REG *)OALPAtoVA(S3C2440A_BASE_REG_PA_IOPORT, FALSE);
	    #if BOARD_TYPE==BOARD_MINI2440
	    s2440IOP->GPGCON |= (0x3<<0 | 0x3<<3 | 0x3<<5);
	    s2440IOP->GPGUP  |= 0x1<<0 | 0x1<<3 | 0x1 <<5;

	    if (!(s2440IOP->GPGDAT & (0x1<<0))) {
	        while(!(s2440IOP->GPGDAT & (0x1<<0)));
	        KeySelect = 0x0d; // enter lcd menu.
	        break;
	    }

	    #elif BOARD_TYPE==BOARD_P1310
	    s2440IOP->GPBUP  |= (0x1<<8 | 0x1<<2 | 0x1<<10);
	    s2440IOP->GPBCON &= ~( 0x3<<16 | 0x3<<4 | 0x2<<20);

        if (!(s2440IOP->GPBDAT & (0x1<<2))) {
            while(!(s2440IOP->GPBDAT & (0x1<<2)));// wait button release...
            KeySelect = 0x0d; // enter lcd menu.
            break;
        }
	    #endif        
}
#endif

		
        dwCurrTime = OEMEthGetSecs();

        if (dwCurrTime > dwPrevTime)
        {
            int i, j;

            // 1 Second has elapsed - update the countdown timer.
            dwPrevTime = dwCurrTime;
            if (BootDelay < 9)
                i = 11;
            else if (BootDelay < 99)
                i = 12;
            else if (BootDelay < 999)
                i = 13;

            for(j = 0; j < i; j++)
                OEMWriteDebugByte((BYTE)0x08); // print back space

#ifdef SUPPORT_LCD_FUNCTION
			DrawNum(STARTX, 16, BootDelay, LCD_RED, LCD_BLACK, NORMAL);
#endif                
            EdbgOutputDebugString ( "%d seconds. ", BootDelay--);
			
        }
    }
	OALMSG(OAL_INFO, (TEXT("\r\n")));

    // Boot or enter bootloader menu.
    //
    switch(KeySelect)
    {
    case 0x20: // Boot menu.
        g_bDownloadImage = MainMenu(g_pBootCfg);
        break;
    case 0x0d: // the user cancelled the countdown.
        g_bDownloadImage = MenuLcd(g_pBootCfg);
        break;
    case 0x00: // Fall through if no keys were pressed -or-
    default:
//        if (g_pBootCfg->ConfigFlags & BOOT_TYPE_DIRECT)
        {
			OALMSG(TRUE, (TEXT("\r\nLaunching image from boot media ... \r\n")));
            g_bDownloadImage = FALSE;
        }
//        else
//        {
//			OALMSG(TRUE, (TEXT("\r\nStarting auto-download ... \r\n")));
//            g_bDownloadImage = TRUE;
//        }
        break;
    }

    if ( !g_bDownloadImage )
    {
        // User doesn't want to download image - load it from the boot media.
        // We could read an entire nk.bin or nk.nb0 into ram and jump.
        if ( !VALID_TOC(g_pTOC) ) {
			OALMSG(OAL_ERROR, (TEXT("OEMPlatformInit: ERROR_INVALID_TOC, can not autoboot.\r\n")));
            return FALSE;
        }
        switch (g_ImageType) {
        	case IMAGE_TYPE_STEPLDR:
				OALMSG(TRUE, (TEXT("Don't support launch STEPLDR.bin\r\n")));
                break;
        	
            case IMAGE_TYPE_LOADER:
				OALMSG(TRUE, (TEXT("Don't support launch EBOOT.bin\r\n")));
                break;

            case IMAGE_TYPE_RAMIMAGE:
				OALMSG(TRUE, (TEXT("OEMPlatformInit: IMAGE_TYPE_RAMIMAGE\r\n")));
                if ( !ReadOSImageFromBootMedia( ) ) 
                {
                    #ifdef SUPPORT_LCD_FUNCTION
    				DrawString(STARTX, LINE(14), "OEMPlatformInit ERROR: Failed to load kernel region into RAM.", LCD_RED, LCD_WHITE, NORMAL);
                    #endif
                    OALMSG(OAL_ERROR, (TEXT("OEMPlatformInit ERROR: Failed to load kernel region into RAM.\r\n")));
                    return FALSE;
                }
				break;

            default:
				OALMSG(OAL_ERROR, (TEXT("OEMPlatformInit ERROR: unknown image type: 0x%x \r\n"), g_ImageType));
                return FALSE;
        }
    }

    // Configure Ethernet controller.
    //
//	if ( g_bDownloadImage && (g_bUSBDownload == FALSE) )    
	if ( g_bDownloadImage && (g_bUSBDownload==FALSE) && (g_bReadFromSDCard==FALSE))
	{
		if (!InitEthDevice(g_pBootCfg))
    	{
        	OALMSG(OAL_ERROR, (TEXT("ERROR: OEMPlatformInit: Failed to initialize Ethernet controller.\r\n")));
        	goto CleanUp;
        }
	}

    bResult = TRUE;

CleanUp:

    OALMSG(OAL_FUNC, (TEXT("_OEMPlatformInit.\r\n")));
    return(bResult);
}


/*
    @func   DWORD | OEMPreDownload | Complete pre-download tasks - get IP address, initialize TFTP, etc.
    @rdesc  BL_DOWNLOAD = Platform Builder is asking us to download an image, BL_JUMP = Platform Builder is requesting we jump to an existing image, BL_ERROR = Failure.
    @comm    
    @xref   
*/
DWORD OEMPreDownload(void)
{
    BOOL  bGotJump = FALSE;
    DWORD dwDHCPLeaseTime = 0;
    PDWORD pdwDHCPLeaseTime = &dwDHCPLeaseTime;
    DWORD dwBootFlags = 0;

    OALMSG(OAL_FUNC, (TEXT("+OEMPreDownload.\r\n")));

    // Create device name based on Ethernet address (this is how Platform Builder identifies this device).
    //
    OALKitlCreateName(BSP_DEVICE_PREFIX, pBSPArgs->kitl.mac, pBSPArgs->deviceId);
    OALMSG(OAL_INFO, (L"INFO: *** Device Name '%hs' ***\r\n", pBSPArgs->deviceId));

	if  (g_bReadFromSDCard) {
		OALMSG(1, (L"Reading from SD CARD.\r\n"));
	}		
	else if ( (g_bUSBDownload == FALSE ))
	{
		// If the user wants to use a static IP address, don't request an address 
		// from a DHCP server.  This is done by passing in a NULL for the DHCP 
		// lease time variable.  If user specified a static IP address, use it (don't use DHCP).
		//
		if (!(g_pBootCfg->ConfigFlags & CONFIG_FLAGS_DHCP))
		{
			// Static IP address.
			pBSPArgs->kitl.ipAddress  = g_pBootCfg->EdbgAddr.dwIP;
			pBSPArgs->kitl.ipMask     = g_pBootCfg->SubnetMask;
			pBSPArgs->kitl.flags     &= ~OAL_KITL_FLAGS_DHCP;
			pdwDHCPLeaseTime = NULL;
			OALMSG(OAL_INFO, (TEXT("INFO: Using static IP address %s.\r\n"), inet_ntoa(pBSPArgs->kitl.ipAddress))); 
			OALMSG(OAL_INFO, (TEXT("INFO: Using subnet mask %s.\r\n"),       inet_ntoa(pBSPArgs->kitl.ipMask))); 
		}
		else
		{
			pBSPArgs->kitl.ipAddress = 0;
			pBSPArgs->kitl.ipMask    = 0;
		}

		if ( !g_bDownloadImage)
		{
			return(BL_JUMP);
		}

		// Initialize the the TFTP transport.
		//
		g_DeviceAddr.dwIP = pBSPArgs->kitl.ipAddress;
		memcpy(g_DeviceAddr.wMAC, pBSPArgs->kitl.mac, (3 * sizeof(UINT16)));
		g_DeviceAddr.wPort = 0;

		if (!EbootInitEtherTransport(&g_DeviceAddr,
									 &pBSPArgs->kitl.ipMask,
									 &bGotJump,
									 pdwDHCPLeaseTime,
									 EBOOT_VERSION_MAJOR,
									 EBOOT_VERSION_MINOR,
									 BSP_DEVICE_PREFIX,
									 pBSPArgs->deviceId,
									 EDBG_CPU_ARM720,
									 dwBootFlags))
		{
			OALMSG(OAL_ERROR, (TEXT("ERROR: OEMPreDownload: Failed to initialize Ethernet connection.\r\n")));
			return(BL_ERROR);
		}


		// If the user wanted a DHCP address, we presumably have it now - save it for the OS to use.
		//
		if (g_pBootCfg->ConfigFlags & CONFIG_FLAGS_DHCP)
		{
			// DHCP address.
			pBSPArgs->kitl.ipAddress  = g_DeviceAddr.dwIP;
			pBSPArgs->kitl.flags     |= OAL_KITL_FLAGS_DHCP;
		}
        
		OALMSG(OAL_FUNC, (TEXT("_OEMPreDownload.\r\n")));
	}
	else
	{
	    DWORD addr;
        DWORD size;
		OALMSG(TRUE, (TEXT("Please send the Image through USB.\r\n"))); 
//chen yongqiang. dnw.dnwᷢ8ֽ,ebootǶ.
        OEMReadData(4, (LPBYTE)&addr);
        OEMReadData(4, (LPBYTE)&size);
		ImageSize = size;
		OALMSG(TRUE, (TEXT("address:0x%X, size:0x%X\r\n"), addr, size));
	}

    return(bGotJump ? BL_JUMP : BL_DOWNLOAD);
}



BOOL SDCardReadData(DWORD cbData, LPBYTE pbData)
{
	if (cbData == fatRead(pbData, cbData))
		return TRUE;
	else 
		return FALSE;
}




/*
    @func   BOOL | OEMReadData | Generically read download data (abstracts actual transport read call).
    @rdesc  TRUE = Success, FALSE = Failure.
    @comm    
    @xref   
*/
BOOL OEMReadData(DWORD dwData, PUCHAR pData)
{
	BOOL ret;
	static DWORD progress = 0;
	
    OALMSG(OAL_FUNC, (TEXT("+OEMReadData.\r\n")));
//	OALMSG(TRUE, (TEXT("\r\nINFO: dwData = 0x%x, pData = 0x%x \r\n"), dwData, pData));

	if (g_bReadFromSDCard == TRUE) {
		ret = SDCardReadData(dwData, pData);
	} else if ( g_bUSBDownload == FALSE )	{
		ret = EbootEtherReadData(dwData, pData);
	} else {
	    RETAILMSG(0, (L"OEMReadData() addr:%x, size:%x\r\n", *pData, dwData));
		ret = UbootReadData(dwData, pData);
	}


	progress += dwData;
	
	if (g_bReadFromSDCard == TRUE) {
#ifdef SUPPORT_LCD_FUNCTION
		DrawString(32, LINE(11), "%%", LCD_WHITE, LCD_BLACK, NORMAL);
		DrawNum(8, LINE(11), progress*100/ImageSize, LCD_RED, LCD_BLACK, NORMAL);
		DrawString(40, LINE(11), " Bytes:", LCD_WHITE, LCD_BLACK, NORMAL);
		DrawNum(96, LINE(11), progress, LCD_RED, LCD_BLACK, NORMAL);
#endif
//		if (progress==0)
//			OALMSG(TRUE, (L"\r\n\r\n\r\n\r\n\r\n[%10ubytes][%%%3u]", progress, 100*progress/ImageSize));
//		else
//			OALMSG(TRUE, (L"\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b[%10ubytes][%%%3u]", progress, 100*progress/ImageSize));

		if (progress >= ImageSize)
			progress = 0;

	}	
		
/*
	OALMSG(TRUE, (TEXT("\r\n")));
	for ( i = 0; i < dwData; i++ )
	{
		OALMSG(TRUE, (TEXT("<%x>"), *(pData+i)));
		if ( i % 16 == 15 )
			OALMSG(TRUE, (TEXT("\r\n")));
	}
	OALMSG(TRUE, (TEXT("\r\n")));
*/
	return(ret);
}


/*
    @func   void | OEMShowProgress | Displays download progress for the user.
    @rdesc  N/A.
    @comm    
    @xref   
*/
void OEMShowProgress(DWORD dwPacketNum)
{
	OALMSG(OAL_FUNC, (TEXT("+OEMShowProgress.\r\n")));
}


/*
    @func   void | OEMLaunch | Executes the stored/downloaded image.
    @rdesc  N/A.
    @comm    
    @xref   
*/

void OEMLaunch( DWORD dwImageStart, DWORD dwImageLength, DWORD dwLaunchAddr, const ROMHDR *pRomHdr )
{
    DWORD dwPhysLaunchAddr;
    EDBG_ADDR EshellHostAddr;
    EDBG_OS_CONFIG_DATA *pCfgData;    

    OALMSG(OAL_FUNC, (TEXT("+OEMLaunch.\r\n")));

    // If the user requested that a disk image (stored in RAM now) be written to the SmartMedia card, so it now.
    //
    if (g_bDownloadImage && (g_pBootCfg->ConfigFlags & TARGET_TYPE_NAND))
    {
        // Since this platform only supports RAM images, the image cache address is the same as the image RAM address.
        //

        switch (g_ImageType) 
        {
        	case IMAGE_TYPE_STEPLDR:
		        if (!WriteRawImageToBootMedia(dwImageStart, dwImageLength, dwLaunchAddr))
		        {
            		OALMSG(OAL_ERROR, (TEXT("ERROR: OEMLaunch: Failed to store image to Smart Media.\r\n")));
            		goto CleanUp;
        		}
		        OALMSG(TRUE, (TEXT("INFO: Step loader image stored to Smart Media.  Please Reboot.  Halting...\r\n")));
	        	while(1)
	        	{
            		// Wait...
	        	}
        		break;
        	
            case IMAGE_TYPE_LOADER:
				g_pTOC->id[0].dwLoadAddress = dwImageStart;
				g_pTOC->id[0].dwTtlSectors = FILE_TO_SECTOR_SIZE(dwImageLength);
		        if (!WriteRawImageToBootMedia(dwImageStart, dwImageLength, dwLaunchAddr))
		        {
            		OALMSG(OAL_ERROR, (TEXT("ERROR: OEMLaunch: Failed to store image to Smart Media.\r\n")));
            		goto CleanUp;
        		}
				if (dwLaunchAddr && (g_pTOC->id[0].dwJumpAddress != dwLaunchAddr))
				{
					g_pTOC->id[0].dwJumpAddress = dwLaunchAddr;
					if ( !TOC_Write() ) {
	            		EdbgOutputDebugString("*** OEMLaunch ERROR: TOC_Write failed! Next boot may not load from disk *** \r\n");
					}
	        		TOC_Print();
				}

                // reserved bootloader region. by max.
                if ( g_bBootMediaExist ) {				
                    DWORD i;
                    SectorInfo si;
                    
                    // to keep bootpart off of our reserved blocks we must mark it as bad, reserved & read-only
                    si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
                    si.bBadBlock    = BADBLOCKMARK;
                    si.dwReserved1  = 0xffffffff;
                    si.wReserved2   = 0xffff;

	                OALMSG(TRUE, (TEXT("Reserving Blocks [0x%x - 0x%x] ...\r\n"), 0, g_dwImageStartBlock-1));
	                for (i = 0; i < BLOCK_TO_SECTOR(g_dwImageStartBlock); i++) {
                        FMD_WriteSector(i, NULL, &si, 1);
                    }
                    OALMSG(TRUE, (TEXT("...reserve complete.\r\n")));

                } 

                
		        OALMSG(TRUE, (TEXT("INFO: Eboot image stored to Smart Media.  Please Reboot.  Halting...\r\n")));

                #ifdef SUPPORT_LCD_FUNCTION
				DrawString(STARTX, LINE(15), "EBOOT stored to nand. Press  [OK] to Reboot", LCD_RED, LCD_WHITE, NORMAL);
                #endif				
                
		        while(1)
		        {
            		// Wait...
            		#ifdef SUPPORT_LCD_FUNCTION
            		if (OEMGetKey() == KEY_ENTER)
						SystemPowerOff();
					#endif
        		}        		
        		        		
                break;

            case IMAGE_TYPE_RAMIMAGE:
				g_pTOC->id[g_dwTocEntry].dwLoadAddress = dwImageStart;
				g_pTOC->id[g_dwTocEntry].dwTtlSectors = FILE_TO_SECTOR_SIZE(dwImageLength);
                #ifdef SUPPORT_LCD_FUNCTION
				DrawString(STARTX, LINE(11), "writing... please wait.", LCD_RED, LCD_WHITE, NORMAL);
                #endif                
		        if (!WriteOSImageToBootMedia(dwImageStart, dwImageLength, dwLaunchAddr))
		        {
                    #ifdef SUPPORT_LCD_FUNCTION
    				DrawString(STARTX, LINE(12), "FAILED!", LCD_RED, LCD_WHITE, NORMAL);
                    #endif		        
            		OALMSG(OAL_ERROR, (TEXT("ERROR: OEMLaunch: Failed to store image to Smart Media.\r\n")));
            		goto CleanUp;
        		}

				if (dwLaunchAddr && (g_pTOC->id[g_dwTocEntry].dwJumpAddress != dwLaunchAddr))
				{
					g_pTOC->id[g_dwTocEntry].dwJumpAddress = dwLaunchAddr;
					if ( !TOC_Write() ) {
                        #ifdef SUPPORT_LCD_FUNCTION
        				DrawString(STARTX, LINE(12), "FAILED!", LCD_RED, LCD_WHITE, NORMAL);
                        #endif		        
	            		EdbgOutputDebugString("*** OEMLaunch ERROR: TOC_Write failed! Next boot may not load from disk *** \r\n");
					}
	        		TOC_Print();
				}
				else
				{
					dwLaunchAddr= g_pTOC->id[g_dwTocEntry].dwJumpAddress;
					EdbgOutputDebugString("INFO: using TOC[%d] dwJumpAddress: 0x%x\r\n", g_dwTocEntry, dwLaunchAddr);
				}
        		        		
                break;
        }
    }
    else if(g_bDownloadImage)
    {
        switch (g_ImageType) 
        {
        	case IMAGE_TYPE_STEPLDR:
		        OALMSG(TRUE, (TEXT("Stepldr image can't launch from ram.\r\n")));
		        OALMSG(TRUE, (TEXT("You should program it into flash.\r\n")));
		        SpinForever();
				break;
            case IMAGE_TYPE_LOADER:
		        OALMSG(TRUE, (TEXT("Eboot image can't launch from ram.\r\n")));
		        OALMSG(TRUE, (TEXT("You should program it into flash.\r\n")));
		        SpinForever();
		        break;
            default:
            	break;
    	}
    }

    OALMSG(1, (TEXT("waitforconnect\r\n")));
    // Wait for Platform Builder to connect after the download and send us IP and port settings for service
    // connections - also sends us KITL flags.  This information is used later by the OS (KITL).
    //
    if (~g_bUSBDownload & g_bDownloadImage & g_bWaitForConnect & ~g_bReadFromSDCard)
    {
        memset(&EshellHostAddr, 0, sizeof(EDBG_ADDR));

        g_DeviceAddr.dwIP  = pBSPArgs->kitl.ipAddress;
        memcpy(g_DeviceAddr.wMAC, pBSPArgs->kitl.mac, (3 * sizeof(UINT16)));
        g_DeviceAddr.wPort = 0;

        if (!(pCfgData = EbootWaitForHostConnect(&g_DeviceAddr, &EshellHostAddr)))
        {
            OALMSG(OAL_ERROR, (TEXT("ERROR: OEMLaunch: EbootWaitForHostConnect failed.\r\n")));
            goto CleanUp;
        }

        // If the user selected "passive" KITL (i.e., don't connect to the target at boot time), set the
        // flag in the args structure so the OS image can honor it when it boots.
        //
        if (pCfgData->KitlTransport & KTS_PASSIVE_MODE)
        {
            pBSPArgs->kitl.flags |= OAL_KITL_FLAGS_PASSIVE;
        }
	}
    
    // If a launch address was provided, we must have downloaded the image, save the address in case we
    // want to jump to this image next time.  If no launch address was provided, retrieve the last one.
    //
	if (dwLaunchAddr && (g_pTOC->id[g_dwTocEntry].dwJumpAddress != dwLaunchAddr))
	{
		g_pTOC->id[g_dwTocEntry].dwJumpAddress = dwLaunchAddr;
	}
	else
	{
		dwLaunchAddr= g_pTOC->id[g_dwTocEntry].dwJumpAddress;
		OALMSG(OAL_INFO, (TEXT("INFO: using TOC[%d] dwJumpAddress: 0x%x\r\n"), g_dwTocEntry, dwLaunchAddr));
	}

    // Jump to downloaded image (use the physical address since we'll be turning the MMU off)...
    //
    dwPhysLaunchAddr = (DWORD)OALVAtoPA((void *)dwLaunchAddr);
    OALMSG(TRUE, (TEXT("INFO: OEMLaunch: Jumping to Physical Address 0x%Xh (Virtual Address 0x%Xh)...\r\n\r\n\r\n"), dwPhysLaunchAddr, dwLaunchAddr));

    // Jump...
    //
    Launch(dwPhysLaunchAddr);


CleanUp:

    OALMSG(TRUE, (TEXT("ERROR: OEMLaunch: Halting...\r\n")));
    SpinForever();
}


//------------------------------------------------------------------------------
//
//  Function Name:  OEMVerifyMemory( DWORD dwStartAddr, DWORD dwLength )
//  Description..:  This function verifies the passed address range lies
//                  within a valid region of memory. Additionally this function
//                  sets the g_ImageType if the image is a boot loader. 
//  Inputs.......:  DWORD           Memory start address
//                  DWORD           Memory length
//  Outputs......:  BOOL - true if verified, false otherwise
//
//------------------------------------------------------------------------------

BOOL OEMVerifyMemory( DWORD dwStartAddr, DWORD dwLength )
{

    OALMSG(OAL_FUNC, (TEXT("+OEMVerifyMemory.\r\n")));

    // Is the image being downloaded the stepldr?
    if ((dwStartAddr >= STEPLDR_RAM_IMAGE_BASE) &&
        ((dwStartAddr + dwLength - 1) < (STEPLDR_RAM_IMAGE_BASE + STEPLDR_RAM_IMAGE_SIZE)))
    {
        OALMSG(OAL_INFO, (TEXT("Stepldr image\r\n")));
        g_ImageType = IMAGE_TYPE_STEPLDR;     // Stepldr image.
        return TRUE;
    }
    // Is the image being downloaded the bootloader?
    else if ((dwStartAddr >= EBOOT_STORE_ADDRESS) &&
        ((dwStartAddr + dwLength - 1) < (EBOOT_STORE_ADDRESS + EBOOT_STORE_MAX_LENGTH)))
    {
        OALMSG(OAL_INFO, (TEXT("Eboot image\r\n")));
        g_ImageType = IMAGE_TYPE_LOADER;     // Eboot image.
        return TRUE;
    }

    // Is it a ram image?
    else if ((dwStartAddr >= ROM_RAMIMAGE_START) &&
        ((dwStartAddr + dwLength - 1) < (ROM_RAMIMAGE_START + ROM_RAMIMAGE_SIZE)))
    {
        OALMSG(OAL_INFO, (TEXT("RAM image\r\n")));
        g_ImageType = IMAGE_TYPE_RAMIMAGE;
        return TRUE;
    }
	else if (!dwStartAddr && !dwLength)
	{
        OALMSG(TRUE, (TEXT("Don't support raw image\r\n")));
		g_ImageType = IMAGE_TYPE_RAWBIN;
    	return FALSE;
	}

    // HACKHACK: get around MXIP images with funky addresses
    OALMSG(TRUE, (TEXT("BIN image type unknow\r\n")));

    OALMSG(OAL_FUNC, (TEXT("_OEMVerifyMemory.\r\n")));

    return FALSE;
}

/*
    @func   void | OEMMultiBINNotify | Called by blcommon to nofity the OEM code of the number, size, and location of one or more BIN regions,
                                       this routine collects the information and uses it when temporarily caching a flash image in RAM prior to final storage.
    @rdesc  N/A.
    @comm
    @xref
*/
void OEMMultiBINNotify(const PMultiBINInfo pInfo)
{
    BYTE nCount;
	DWORD g_dwMinImageStart;

    OALMSG(OAL_FUNC, (TEXT("+OEMMultiBINNotify.\r\n")));

    if (!pInfo || !pInfo->dwNumRegions)
    {
        OALMSG(OAL_WARN, (TEXT("WARNING: OEMMultiBINNotify: Invalid BIN region descriptor(s).\r\n")));
        return;
    }

	if (!pInfo->Region[0].dwRegionStart && !pInfo->Region[0].dwRegionLength)
	{
    	return;
	}

    g_dwMinImageStart = pInfo->Region[0].dwRegionStart;

    OALMSG(TRUE, (TEXT("\r\nDownload BIN file information:\r\n")));
    OALMSG(TRUE, (TEXT("-----------------------------------------------------\r\n")));
    for (nCount = 0 ; nCount < pInfo->dwNumRegions ; nCount++)
    {
        OALMSG(TRUE, (TEXT("[%d]: Base Address=0x%x  Length=0x%x\r\n"),
            nCount, pInfo->Region[nCount].dwRegionStart, pInfo->Region[nCount].dwRegionLength));
        if (pInfo->Region[nCount].dwRegionStart < g_dwMinImageStart)
        {
            g_dwMinImageStart = pInfo->Region[nCount].dwRegionStart;
            if (g_dwMinImageStart == 0)
            {
                OALMSG(OAL_WARN, (TEXT("WARNING: OEMMultiBINNotify: Bad start address for region (%d).\r\n"), nCount));
                return;
            }
        }
    }

    memcpy((LPBYTE)&g_BINRegionInfo, (LPBYTE)pInfo, sizeof(MultiBINInfo));

    OALMSG(TRUE, (TEXT("-----------------------------------------------------\r\n")));
    OALMSG(OAL_FUNC, (TEXT("_OEMMultiBINNotify.\r\n")));
}

/////////////////////// START - Stubbed functions - START //////////////////////////////
/*
    @func   void | SC_WriteDebugLED | Write to debug LED.
    @rdesc  N/A.
    @comm    
    @xref   
*/

void SC_WriteDebugLED(USHORT wIndex, ULONG dwPattern)
{
    // Stub - needed by NE2000 EDBG driver...
    //
}


ULONG HalSetBusDataByOffset(IN BUS_DATA_TYPE BusDataType,
                            IN ULONG BusNumber,
                            IN ULONG SlotNumber,
                            IN PVOID Buffer,
                            IN ULONG Offset,
                            IN ULONG Length)
{
    return(0);
}


ULONG
HalGetBusDataByOffset(IN BUS_DATA_TYPE BusDataType,
                      IN ULONG BusNumber,
                      IN ULONG SlotNumber,
                      IN PVOID Buffer,
                      IN ULONG Offset,
                      IN ULONG Length)
{
    return(0);
}


BOOLEAN HalTranslateBusAddress(IN INTERFACE_TYPE  InterfaceType,
                               IN ULONG BusNumber,
                               IN PHYSICAL_ADDRESS BusAddress,
                               IN OUT PULONG AddressSpace,
                               OUT PPHYSICAL_ADDRESS TranslatedAddress)
{

    // All accesses on this platform are memory accesses...
    //
    if (AddressSpace)
        *AddressSpace = 0;
 
    // 1:1 mapping...
    //
    if (TranslatedAddress)
    {
        *TranslatedAddress = BusAddress;
        return(TRUE);
    }

    return(FALSE);
}


PVOID MmMapIoSpace(IN PHYSICAL_ADDRESS PhysicalAddress,
                   IN ULONG NumberOfBytes,
                   IN BOOLEAN CacheEnable)
{
    DWORD dwAddr = PhysicalAddress.LowPart;

    if (CacheEnable)
        dwAddr &= ~CACHED_TO_UNCACHED_OFFSET; 
    else
        dwAddr |= CACHED_TO_UNCACHED_OFFSET; 

    return((PVOID)dwAddr);
}


VOID MmUnmapIoSpace(IN PVOID BaseAddress,
                    IN ULONG NumberOfBytes)
{
}

VOID WINAPI SetLastError(DWORD dwErrCode)
{
}
/////////////////////// END - Stubbed functions - END //////////////////////////////


/*
    @func   PVOID | GetKernelExtPointer | Locates the kernel region's extension area pointer.
    @rdesc  Pointer to the kernel's extension area.
    @comm    
    @xref   
*/
PVOID GetKernelExtPointer(DWORD dwRegionStart, DWORD dwRegionLength)
{
    DWORD dwCacheAddress = 0;
    ROMHDR *pROMHeader;
    DWORD dwNumModules = 0;
    TOCentry *pTOC;

    if (dwRegionStart == 0 || dwRegionLength == 0)
        return(NULL);

    if (*(LPDWORD) OEMMapMemAddr (dwRegionStart, dwRegionStart + ROM_SIGNATURE_OFFSET) != ROM_SIGNATURE)
        return NULL;

    // A pointer to the ROMHDR structure lives just past the ROM_SIGNATURE (which is a longword value).  Note that
    // this pointer is remapped since it might be a flash address (image destined for flash), but is actually cached
    // in RAM.
    //
    dwCacheAddress = *(LPDWORD) OEMMapMemAddr (dwRegionStart, dwRegionStart + ROM_SIGNATURE_OFFSET + sizeof(ULONG));
    pROMHeader     = (ROMHDR *) OEMMapMemAddr (dwRegionStart, dwCacheAddress);

    // Make sure sure are some modules in the table of contents.
    //
    if ((dwNumModules = pROMHeader->nummods) == 0)
        return NULL;

    // Locate the table of contents and search for the kernel executable and the TOC immediately follows the ROMHDR.
    //
    pTOC = (TOCentry *)(pROMHeader + 1);

    while(dwNumModules--) {
        LPBYTE pFileName = OEMMapMemAddr(dwRegionStart, (DWORD)pTOC->lpszFileName);
        if (!strcmp((const char *)pFileName, "nk.exe")) {
            return ((PVOID)(pROMHeader->pExtensions));
        }
        ++pTOC;
    }
    return NULL;
}


/*
    @func   BOOL | OEMDebugInit | Initializes the serial port for debug output message.
    @rdesc  TRUE == Success and FALSE == Failure.
    @comm    
    @xref   
*/
BOOL OEMDebugInit(void)
{

    // Set up function callbacks used by blcommon.
    //
    g_pOEMVerifyMemory   = OEMVerifyMemory;      // Verify RAM.
    g_pOEMMultiBINNotify = OEMMultiBINNotify;

    // Call serial initialization routine (shared with the OAL).
    //
    OEMInitDebugSerial();

    return(TRUE);
}

#ifdef SUPPORT_LCD_FUNCTION


BYTE OEMGetKey(void)
{
    BYTE    key = 0xFF;
    volatile DWORD t;
    
    volatile S3C2440A_IOPORT_REG *s2440IOP = (S3C2440A_IOPORT_REG *)OALPAtoVA(S3C2440A_BASE_REG_PA_IOPORT, FALSE);
#if BOARD_TYPE==BOARD_MINI2440
    while(TRUE) {
        if (!(s2440IOP->GPGDAT & (0x1<<3))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPGDAT & (0x1<<3)));
            key = KEY_UP; break;
        }    
        if (!(s2440IOP->GPGDAT & (0x1<<5))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPGDAT & (0x1<<5)));
            key = KEY_DOWN; break;
        }    
        if (!(s2440IOP->GPGDAT & (0x1<<0))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPGDAT & (0x1<<0)));
            key = KEY_ENTER; break;
        }    
        
    }
#elif BOARD_TYPE==BOARD_QQ2440
    while(TRUE) {
        if (!(s2440IOP->GPFDAT & (0x1<<0))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPFDAT & (0x1<<0)));
            key = KEY_UP; break;
        }    
        if (!(s2440IOP->GPFDAT & (0x1<<2))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPFDAT & (0x1<<2)));
            key = KEY_DOWN; break;
        }    
        if (!(s2440IOP->GPGDAT & (0x1<<3))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPGDAT & (0x1<<3)));
            key = KEY_ENTER; break;
        }    
        
    }
#elif BOARD_TYPE==BOARD_P1310
    while(TRUE) {
        if (!(s2440IOP->GPBDAT & (0x1<<8))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPBDAT & (0x1<<8)));
            key = KEY_UP; break;
        }    
        if (!(s2440IOP->GPBDAT & (0x1<<10))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPBDAT & (0x1<<10)));
            key = KEY_DOWN; break;
        }    
        if (!(s2440IOP->GPBDAT & (0x1<<2))) {
            for(t=100000;t;t--);
            while(!(s2440IOP->GPBDAT & (0x1<<2)));
            key = KEY_ENTER; break;
        }    
        
    }


#endif
    return key;
}


static BOOL MenuLcd(PBOOT_CFG pBootCfg)
{
    BYTE    Key=0, Sel=0, NeedSave=0;
    BOOL bConfigChanged = FALSE;
    BOOLEAN bDownload = TRUE;

    RETAILMSG(1, (L"MenuLCD(): enter BIOS.\r\n"));
    NeedSave = (pBootCfg->ConfigFlags & TARGET_TYPE_NAND)?1:0;
loop:    
    
	DispOneColor(LCD_BLACK);
    DrawString(0, 0, "...BIOS...090227", LCD_WHITE, LCD_BLACK, TRANSP);
loop_no_clr:    
    DrawString(STARTX, 16*(Sel+2), "                            ", LCD_BLUE, LCD_BLUE, NORMAL);
    
    if (!NeedSave)
        DrawString(STARTX, LINE(0), "Save Image to Nand [NO]", LCD_WHITE, LCD_BLACK, TRANSP);
    else
        DrawString(STARTX, LINE(0), "Save Image to Nand [YES]", LCD_WHITE, LCD_BLACK, TRANSP);         

    DrawString(STARTX, LINE(1), "Down BIOS  From SD card", LCD_WHITE, LCD_BLACK, TRANSP);        
    DrawString(STARTX, LINE(2), "Down NK    From SD card", LCD_WHITE, LCD_BLACK, TRANSP);            
    DrawString(STARTX, LINE(3), "Down LOGO  From SD card  ", LCD_WHITE, LCD_BLACK, TRANSP);    
    DrawString(STARTX, LINE(4), "Down bin   From USB    ", LCD_WHITE, LCD_BLACK, TRANSP);    
    DrawString(STARTX, LINE(5), "Clear Register", LCD_WHITE, LCD_BLACK, TRANSP);        
    DrawString(STARTX, LINE(6), "Format Nand ", LCD_WHITE, LCD_BLACK, TRANSP);
    DrawString(STARTX, LINE(7), "Load image From Nand   ", LCD_WHITE, LCD_BLACK, TRANSP);    	
    
    Key = 0;

    while(!(Key == KEY_UP || Key == KEY_DOWN || Key == KEY_ENTER)) {
        //Key = OEMReadDebugByte();
        Key = OEMGetKey();
    }

    
    if (Key == KEY_UP) { //key UP
    
        Sel = (Sel==0)?(LASTLINE):(Sel-1);
        goto loop;
        
    } else if (Key == KEY_DOWN) { // key DOWN
    
        Sel = (Sel<LASTLINE)?(Sel+1):0;
        goto loop;
        
    } else if (Key == KEY_ENTER) { // key Enter

		DrawString(STARTX, 16*(Sel+2), "                         ==>", LCD_RED, LCD_RED, TRANSP);
	
        if (Sel == 0) {             // enable write nand?
            NeedSave = NeedSave?0:1;
            //pBootCfg->ConfigFlags = (pBootCfg->ConfigFlags ^ TARGET_TYPE_NAND);
            if (NeedSave)
                pBootCfg->ConfigFlags |= TARGET_TYPE_NAND;
            else
                pBootCfg->ConfigFlags &= ~TARGET_TYPE_NAND;
            
            bConfigChanged = TRUE;
            goto loop;
        }
        if (Sel == 4) {             // download from USB        
            bConfigChanged = TRUE;
			g_bUSBDownload = TRUE;
            bDownload = TRUE;
        }
        if (Sel == 1 || Sel == 2 || Sel == 3) { // download from SD card.
			LONG ret;
			ret = SDInit();
			if (ret == SD_API_STATUS_DEVICE_NOT_CONNECTED) {
				DrawString(STARTX, LINE(7), "[SD] card not found", LCD_RED, LCD_BLACK, NORMAL);
			} else if (ret) {
				DrawString(STARTX, LINE(7), "[SD] fail to init ", LCD_RED, LCD_BLACK, NORMAL);
			} else {
				DrawString(STARTX, LINE(7), "[SD] init done     ", LCD_BLACK, LCD_WHITE, NORMAL);	// ok		
			}
		
			if (!fatLoad(0)){
				DrawString(STARTX, LINE(8), "[FAT] fail to read", LCD_RED, LCD_BLACK, NORMAL);
				goto loop_no_clr;  //break;
			}else DrawString(STARTX, LINE(8), "[FAT] mounted     ", LCD_BLACK, LCD_WHITE, NORMAL); // ok
			
			if (Sel==1)   ImageSize = fatOpen("eboot.bin");//file name is not unicode...
			else if (Sel==2)   ImageSize = fatOpen("nk.bin");
            else if (Sel==3)   ImageSize = fatOpen("logo.bin");
            
			if (ImageSize == -1){
				DrawString(STARTX, LINE(9), "[FILE] bin not found      ", LCD_RED, LCD_BLACK, NORMAL);				
				goto loop_no_clr; //break;  
			}else DrawString(STARTX, LINE(9), "[FILE] opened. wait reading", LCD_BLACK, LCD_WHITE, NORMAL); // ok

			//fatRead((void*)0x80200000, ImageSize);
#if 1
            if (Sel == 3){ // read logo from sd card.
                DWORD   i;
                SectorInfo si;

                SDCardReadData(ImageSize, (LPBYTE)IMAGE_FRAMEBUFFER_UA_BASE);
                for (i=LOGO_BLOCK; i<LOGO_BLOCK+LOGO_BLOCK_SIZE; i++) {
                    RETAILMSG(1, (L"write logo to block:%u.\r\n", i));
                    FMD_EraseBlock(i);
                    if (!WriteBlock(i, IMAGE_FRAMEBUFFER_UA_BASE + wBLOCK_SIZE*(i-LOGO_BLOCK), NULL))
                        RETAILMSG(1, (L"failed to write logo to block:%u.", i));
                }

                si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
                si.bBadBlock    = BADBLOCKMARK;
                si.dwReserved1  = 0xffffffff;
                si.wReserved2   = 0xffff;

                for (i = 0; i < BLOCK_TO_SECTOR(g_dwImageStartBlock); i++) {
                    FMD_WriteSector(i, NULL, &si, 1);
                }
          
                DrawString(STARTX, LINE(10), "write logo complete.", LCD_RED, LCD_BLACK, NORMAL);
                goto loop_no_clr;
            }
#else
            if (Sel == 3){ // read logo from sd card.
                DrawString(STARTX, LINE(10), "not support.", LCD_RED, LCD_BLACK, NORMAL);            
                goto loop_no_clr;
            }
#endif
			g_bReadFromSDCard	= TRUE;
			bConfigChanged = TRUE;
			bDownload = TRUE;
            

        }
		

        if (Sel == 6) {	// format nand
            if ( !g_bBootMediaExist ) {
				DrawString(STARTX, LINE(7), "nand device not found", LCD_RED, LCD_BLACK, NORMAL);
            } else {
                SectorInfo si;
                DWORD i;

                // to keep bootpart off of our reserved blocks we must mark it as bad, reserved & read-only
                si.bOEMReserved = OEM_BLOCK_RESERVED | OEM_BLOCK_READONLY;
                si.bBadBlock    = BADBLOCKMARK;
                si.dwReserved1  = 0xffffffff;
                si.wReserved2   = 0xffff;

				// Ҫbootл.
				DrawString(STARTX, LINE(8), "Reserving Blocks...", LCD_RED, LCD_BLACK, NORMAL);
                for (i = 0; i < BLOCK_TO_SECTOR(g_dwImageStartBlock); i++) {
                    FMD_WriteSector(i, NULL, &si, 1);
                }
				DrawString(STARTX, LINE(9), "...reserve complete.", LCD_RED, LCD_BLACK, NORMAL);                

				DrawString(STARTX, LINE(10), "Low-level format Blocks...", LCD_RED, LCD_BLACK, NORMAL);    
                for (i = g_dwImageStartBlock; i < wNUM_BLOCKS; i++) {
                    FMD_EraseBlock(i);
                }
				DrawString(STARTX, LINE(11), "...erase complete.", LCD_RED, LCD_BLACK, NORMAL);    
            }
            goto loop_no_clr;
        }    

        if (Sel == 7) { // load ce image from nand
            bDownload = FALSE;
        }

        if (Sel == 5) { // boot with clear register.
            bDownload = FALSE;
            pBSPArgs->bSysHiveClean = TRUE;
            pBSPArgs->bUserHiveClean = TRUE;
        }

    }


    
    // If eboot settings were changed by user, save them to flash.
    //
    if (bConfigChanged && !TOC_Write())
    {
        OALMSG(OAL_WARN, (TEXT("WARNING: MainMenu: Failed to store updated bootloader configuration to flash.\r\n")));
    }

    return(bDownload);    
    
}

#endif

